class EndPoints:
    BASE = "https://api.aladhan.com/v1/"
    TIMINGS = BASE + "timings"
    TIMINGS_BY_ADDRESS = TIMINGS + "ByAddress"
    TIMINGS_BY_CITY = TIMINGS + "ByCity"
    CALENDAR = BASE + "calendar"
    CALENDAR_BY_ADDRESS = CALENDAR + "ByAddress"
    CALENDAR_BY_CITY = CALENDAR + "ByCity"
    HIJRI_CALENDAR = BASE + "hijriCalendar"
    HIJRI_CALENDAR_BY_ADDRESS = HIJRI_CALENDAR + "ByAddress"
    HIJRI_CALENDAR_BY_CITY = HIJRI_CALENDAR + "ByCity"
    METHODS = BASE + "methods"
    ASMA_AL_HUSNA = BASE + "asmaAlHusna"
    QIBLA = BASE + "qibla"
